/* 
 * Tux Racer 
 * Copyright (C) 1999-2001 Jasmin F. Patry
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "tuxracer.h"
#include "racing.h"
#include "audio.h"
#include "keyboard.h"
#include "keyframe.h"
#include "course_render.h"
#include "multiplayer.h"
#include "gl_util.h"
#include "fps.h"
#include "loop.h"
#include "render_util.h"
#include "view.h"
#include "tux.h"
#include "tux_shadow.h"
#include "phys_sim.h"
#include "part_sys.h"
#include "screenshot.h"
#include "fog.h"
#include "viewfrustum.h"
#include "track_marks.h"
#include "hud.h"
#include "hud_training.h"
#include "joystick.h"
#include "bonus.h"
#include "accelerometer.h"

/* Time constant for automatic steering centering (s) */
#define TURN_DECAY_TIME_CONSTANT 0.5

/* Maximum time over which jump can be charged */
#define MAX_JUMP_AMT 1.0

/* Time constant for automatic rolling centering (s) */
#define ROLL_DECAY_TIME_CONSTANT 0.2

#define JUMP_CHARGE_DECAY_TIME_CONSTANT 0.1

/* If too high off the ground, tux flaps instead of jumping */
#define JUMP_MAX_START_HEIGHT 0.30

/* factor limit to be considered as a left or a right turn. useful only for tricks */
#define TURN_FACTOR_LIMIT 0.1

static bool_t right_turn;
static bool_t left_turn;
static bool_t trick_modifier;
static bool_t paddling;
static bool_t charging;
static bool_t braking;
static scalar_t charge_start_time;
static int last_terrain;

bool_t get_trick_modifier() {
    return trick_modifier;
}

/* ce init ne ser que pendant le tutorial, quand on souhaite faire revenir tux en arriere */
void racing_init_for_tutorial(point_t point) {
    player_data_t *plyr = get_player_data( local_player() );
    plyr->control.is_flying=False;
    plyr->control.fly_total_time=0;
    init_physical_simulation_at_point(plyr,point);
}

void racing_init(void) 
{
    player_data_t *plyr = get_player_data( local_player() );
    
    winsys_set_display_func( main_loop );
    winsys_set_idle_func( main_loop );
    winsys_set_reshape_func( reshape );
    winsys_set_mouse_func( racing_mouse_func );
    winsys_set_motion_func( NULL );
    winsys_set_passive_motion_func( NULL );
        
    /* Initialize view */
    if ( getparam_view_mode() < 0 || 
        getparam_view_mode() >= NUM_VIEW_MODES ) 
    {
        setparam_view_mode( FOLLOW );
    }
    
    if (g_game.practicing) {
        set_view_mode( plyr, (view_mode_t)getparam_view_mode() );
    } else {
    /* en mode training il faut qu'on voie tux pour voir quand il paddle, quand il fait des tricks, etc... */
        set_view_mode( plyr, FOLLOW );
    }
    
    /* We need to reset controls here since callbacks won't have been
     called in paused mode. This results in duplication between this
     code and init_physical_simulation.  Oh well. */
    left_turn = right_turn = paddling = False;
    trick_modifier = False;
    charging = False;
    plyr->control.turn_fact = 0.0;
    plyr->control.turn_animation = 0.0;
    plyr->control.is_braking = False;
    plyr->control.is_paddling = False;
    plyr->control.jumping = False;
    plyr->control.jump_charging = False;
    
#ifdef __APPLE__
    plyr->control.is_flying=False;
    plyr->control.fly_total_time=0;
#endif
    
    /* Set last_terrain to a value not used below */
    last_terrain = 0;
    
    if ( g_game.prev_mode != PAUSED ) {
        init_physical_simulation();
    }
    
    last_terrain = 0;
    
    g_game.race_aborted = False;
#ifdef __APPLE__
    g_game.race_time_over = False;
#endif
    
    play_music( "racing" );
    /* play_sound( "start_race", 0 ); */
}

void calc_jump_amt( scalar_t time_step )
{
    player_data_t *plyr = get_player_data( local_player() );
    
    if ( plyr->control.jump_charging ) {
        plyr->control.jump_amt = min( 
                                     MAX_JUMP_AMT, g_game.time - charge_start_time );
    } else if ( plyr->control.jumping ) {
        plyr->control.jump_amt *= 
	    ( 1.0 - ( g_game.time - plyr->control.jump_start_time ) / 
         JUMP_FORCE_DURATION );
    } else {
        plyr->control.jump_amt = 0;
    }
}

void racing_loop( scalar_t time_step )
{
    int width, height;
    player_data_t *plyr = get_player_data( local_player() );
    bool_t joy_left_turn = False;
    bool_t joy_right_turn = False;
    scalar_t joy_turn_fact = 0.0;
    bool_t joy_paddling = False;
    bool_t joy_braking = False;
    bool_t joy_charging = False;
    bool_t airborne;
    vector_t dir;
    scalar_t speed;
    scalar_t terrain_weights[NumTerrains];
    int new_terrain = 0;
    int slide_volume;
    int ax,ay,az;

    
    dir = plyr->vel;
    speed = normalize_vector(&dir);
    
    airborne = plyr->airborne;
    
    width = getparam_x_resolution();
    height = getparam_y_resolution();
    
    check_gl_error();
    
    new_frame_for_fps_calc();
    
    update_audio();
    
    clear_rendering_context();
    
    setup_fog();
    
    /* Update braking */
    plyr->control.is_braking = (bool_t) ( braking || joy_braking );
    
    if ( airborne ) {
        new_terrain = (1<<NumTerrains);
        
        /*
         * Tricks
         */
        if ( trick_modifier) {
            if (left_turn) {
                plyr->control.barrel_roll_left = True;
            }

            if (right_turn) {
                plyr->control.barrel_roll_right = True;
            }

            if (paddling) {
                plyr->control.front_flip = True;
            }

            if (plyr->control.is_braking) {
                plyr->control.back_flip = True;
            }
            
            //Par défaut ca fait un front flip
            if(!plyr->control.front_flip && !plyr->control.back_flip &&
               !plyr->control.barrel_roll_right && !plyr->control.barrel_roll_left )
            {
                plyr->control.back_flip = True;
            }
#ifdef __APPLE__
            TRDebugLog("tricks : %d",plyr->tricks);
#endif
        }
        
        
    } else {
        
        get_surface_type(plyr->pos.x, plyr->pos.z, terrain_weights);
        if (terrain_weights[Snow] > 0) {
            new_terrain |= (1<<Snow);
        }
        if (terrain_weights[Rock] > 0) {
            new_terrain |= (1<<Rock);
        } 
        if (terrain_weights[Ice] > 0) {
            new_terrain |= (1<<Ice);
        }
        
    }
    
    /*
     * Jumping
     */
    calc_jump_amt( time_step );
    
    if ( ( charging || joy_charging ) && 
        !plyr->control.jump_charging && !plyr->control.jumping ) 
    {
        plyr->control.jump_charging = True;
        charge_start_time = g_game.time;
    }
    
    if ( ( !charging && !joy_charging ) && plyr->control.jump_charging ) {
        plyr->control.jump_charging = False;
        plyr->control.begin_jump = True;
    }
    
    
    /* 
     * Turning 
     */
    if ( ( left_turn || joy_left_turn )  ^ (right_turn || joy_right_turn ) ) {
        bool_t turning_left = (bool_t) ( left_turn || joy_left_turn );

        if ( joy_left_turn || joy_right_turn ) {
            plyr->control.turn_fact = joy_turn_fact;
        } else {
            plyr->control.turn_fact = (turning_left?-1:1);
        }

        plyr->control.turn_animation += (turning_left?-1:1) *
            0.15 * time_step / 0.05;
        plyr->control.turn_animation = 
            min(1.0, max(-1.0, plyr->control.turn_animation));
    } else {
        plyr->control.turn_fact = 0;

        /* Decay turn animation */
        if ( time_step < ROLL_DECAY_TIME_CONSTANT ) {
            plyr->control.turn_animation *= 
                1.0 - time_step/ROLL_DECAY_TIME_CONSTANT;
        } else {
            plyr->control.turn_animation = 0.0;
        }
    }

    accel_read(&ax,&ay,&az);
    scalar_t iPhone_turn_fact=0; //MV accelerometerTurnFact();
    iPhone_turn_fact=-ax/500.0;
    
    /*left_turn and right_turn are informations useful for tricks*/
    if (iPhone_turn_fact>TURN_FACTOR_LIMIT) {
        left_turn=false;
        right_turn=true;
    }
    else if (iPhone_turn_fact<-TURN_FACTOR_LIMIT) {
        left_turn=true;
        right_turn=false;
    } else left_turn = right_turn = false;
    
    plyr->control.turn_fact = iPhone_turn_fact;
    plyr->control.turn_animation = iPhone_turn_fact;
    
    
    /*
     * Paddling
     */
    if ( ( paddling || joy_paddling ) && plyr->control.is_paddling == False ) {
        plyr->control.is_paddling = True;
        plyr->control.paddle_time = g_game.time;
    }
    
    /*
     * Play flying sound (__APPLE__ : and add Flying time to plyr->control.fly_total_time)
     */
    if (new_terrain & (1<<NumTerrains)) {
        set_sound_volume("flying_sound", min(128, speed*2));
        
        if (!(last_terrain & (1<<NumTerrains))) {
            play_sound( "flying_sound", -1 );
            plyr->control.is_flying=true;
            plyr->control.fly_start_time = g_game.time;
        }
    } else {
        if (last_terrain & (1<<NumTerrains)) {
            plyr->control.is_flying=false;
            plyr->control.fly_end_time = g_game.time;
            if (plyr->control.fly_end_time-plyr->control.fly_start_time>FLYING_TIME_LIMIT) {
                plyr->control.fly_total_time += plyr->control.fly_end_time-plyr->control.fly_start_time;
            }
            halt_sound( "flying_sound" );
        }
    }
    
    /*
     * Play sliding sound
     */
    slide_volume = min( (((pow(plyr->control.turn_fact, 2)*128)) +
                         (plyr->control.is_braking?128:0) +
                         (plyr->control.jumping?128:0) +
                         20) *
                       (speed/10), 128 );
    if (new_terrain & (1<<Snow)) {
        set_sound_volume("snow_sound", slide_volume * terrain_weights[Snow]);
        if (!(last_terrain & (1<<Snow))) {
            play_sound( "snow_sound", -1 );
        }
    } else {
        if (last_terrain & (1<<Snow)) {
            halt_sound( "snow_sound" );
        }
    }
    if (new_terrain & (1<<Rock)) {
        set_sound_volume("rock_sound", 128*pow((speed/2), 2) * 
                         terrain_weights[Rock]);
        if (!(last_terrain & (1<<Rock))) {
            play_sound( "rock_sound", -1 );
        }
    } else {
        if (last_terrain & (1<<Rock)) {
            halt_sound( "rock_sound" );
        }
    }
    if (new_terrain & (1<<Ice)) {
        set_sound_volume("ice_sound", slide_volume * terrain_weights[Ice]);
        if (!(last_terrain & (1<<Ice))) {
            play_sound( "ice_sound", -1 );
        }
    } else {
        if (last_terrain & (1<<Ice)) {
            halt_sound( "ice_sound" );
        }
    }
    last_terrain = new_terrain; 
    
    
    /*
     * gs
     */
    bool roll = plyr->control.barrel_roll_left || plyr->control.barrel_roll_right;
    bool flip = plyr->control.front_flip || plyr->control.back_flip;
    if(roll && plyr->control.barrel_roll_factor == 0)
    {
        if(flip)
        {
            if(plyr->control.barrel_roll_left)
                add_new_bonus("Hyper heavy Jump", get_score_for_trick(HYPER_HEAVY_JUMP));
            else
                add_new_bonus("Ray star hybrid Jump", get_score_for_trick(RAY_STAR_HYBRID_JUMP));
        }
        else
        {
            if(plyr->control.barrel_roll_left)
                add_new_bonus("Roll Left", get_score_for_trick(ROLL_LEFT));
            else
                add_new_bonus("Roll Right", get_score_for_trick(ROLL_RIGHT));
        }
    }

    if(flip && plyr->control.flip_factor == 0)
    {
        if(roll)
        {
            if(plyr->control.back_flip)
                add_new_bonus("Saturn ice Fever", get_score_for_trick(SATURN_ICE_FEVER));
            else
                add_new_bonus("Wild pinguin Show", get_score_for_trick(WILD_PINGUIN_SHOW));
        }
        else
        {
            if(plyr->control.back_flip)
                add_new_bonus("Back Flip", get_score_for_trick(BACK_FLIP));
            else
                add_new_bonus("Barlow's Wheel", get_score_for_trick(BARLOWS_WHEEL));
        }
    }

    if (roll) {
        plyr->tricks+=1;
        plyr->control.barrel_roll_factor += 
		( plyr->control.barrel_roll_left ? -1 : 1 ) * 0.05 * time_step / 0.05;
        if ( (plyr->control.barrel_roll_factor  > 1) ||
            (plyr->control.barrel_roll_factor  < -1) ) {
            plyr->control.barrel_roll_factor = 0;
            plyr->control.barrel_roll_left = plyr->control.barrel_roll_right = False;
        }
    }
    if (flip) {
        plyr->tricks+=1;
        plyr->control.flip_factor += 
		( plyr->control.back_flip ? -1 : 1 ) * 0.05 * time_step / 0.05;
        if ( (plyr->control.flip_factor  > 1) ||
            (plyr->control.flip_factor  < -1) ) {
            plyr->control.flip_factor = 0;
            plyr->control.front_flip = plyr->control.back_flip = False;
        }
    }

    update_player_pos( plyr, time_step );
	
    /* 
     * Track Marks
     */
    add_track_mark( plyr );
    
    
    update_view( plyr, time_step );
    
    setup_view_frustum( plyr, NEAR_CLIP_DIST, 
                       getparam_forward_clip_distance() );
    
    draw_sky(plyr->view.pos);
    
    draw_fog_plane();
    
    set_course_clipping( True );
    set_course_eye_point( plyr->view.pos );
    setup_course_lighting();
    render_course();
    draw_trees();
    
    if ( getparam_draw_particles() ) {
        update_particles( time_step );
        draw_particles( plyr );
    }
    
    draw_tux();
    draw_tux_shadow();
    
    draw_hud( plyr );
    
    draw_hud_training(plyr);
    
    reshape( width, height );
    
    winsys_swap_buffers();
    
    g_game.time += time_step;
} 

static void racing_term(void)
{
    halt_sound( "flying_sound" );
    halt_sound( "rock_sound" );
    halt_sound( "ice_sound" );
    halt_sound( "snow_sound" );
    break_track_marks();
}


START_KEYBOARD_CB( quit_racing_cb )
{
    if ( release ) return;
    g_game.race_aborted = True;
    set_game_mode( GAME_OVER );
}
END_KEYBOARD_CB


START_KEYBOARD_CB( turn_left_cb )
{
    left_turn = (bool_t) !release;
}
END_KEYBOARD_CB


START_KEYBOARD_CB( turn_right_cb )
{
    right_turn = (bool_t) !release;
}
END_KEYBOARD_CB


START_KEYBOARD_CB( trick_modifier_cb )
{
    trick_modifier = (bool_t) !release;
}
END_KEYBOARD_CB


START_KEYBOARD_CB( brake_cb )
{
    braking = (bool_t) !release;
}
END_KEYBOARD_CB

START_KEYBOARD_CB( paddle_cb )
{
    paddling = (bool_t) !release;
    if (paddling) plyr->control.is_accelerating = true; else  plyr->control.is_accelerating = false;
}
END_KEYBOARD_CB


START_KEYBOARD_CB( above_view_cb )
{
    if ( release ) return;
    set_view_mode( plyr, ABOVE );
    setparam_view_mode( ABOVE );
}
END_KEYBOARD_CB

START_KEYBOARD_CB( follow_view_cb )
{
    if ( release ) return;
    set_view_mode( plyr, FOLLOW );
    setparam_view_mode( FOLLOW );
}
END_KEYBOARD_CB

START_KEYBOARD_CB( behind_view_cb )
{
    if ( release ) return;
    set_view_mode( plyr, BEHIND );
    setparam_view_mode( BEHIND );
}
END_KEYBOARD_CB

START_KEYBOARD_CB( screenshot_cb )
{
    if ( release ) return;
    screenshot();
}
END_KEYBOARD_CB

START_KEYBOARD_CB( pause_cb )
{
    if ( release ) return;
    g_game.race_paused;
    set_game_mode( PAUSED );
}
END_KEYBOARD_CB

START_KEYBOARD_CB( reset_cb )
{
    if ( release ) return;
    set_game_mode( RESET );
}
END_KEYBOARD_CB

START_KEYBOARD_CB( jump_cb )
{
    charging = (bool_t) !release;
}
END_KEYBOARD_CB

void racing_register()
{
    int status = 0;
    
    status |= add_keymap_entry( RACING, CONFIGURABLE_KEY, 
                               "backspace", getparam_quit_key, quit_racing_cb );
    status |= add_keymap_entry( RACING, CONFIGURABLE_KEY, 
                               "j", getparam_turn_left_key, turn_left_cb );
    status |= add_keymap_entry( RACING, CONFIGURABLE_KEY, 
                               "l", getparam_turn_right_key, turn_right_cb );
    status |= add_keymap_entry( RACING, CONFIGURABLE_KEY, 
                               "d", getparam_trick_modifier_key, 
                               trick_modifier_cb );
    status |= add_keymap_entry( RACING, CONFIGURABLE_KEY, 
                               "b", getparam_brake_key, brake_cb );
    status |= add_keymap_entry( RACING, CONFIGURABLE_KEY, 
                               "k", getparam_paddle_key, paddle_cb );
    
    status |= add_keymap_entry( RACING, CONFIGURABLE_KEY, 
                               "z", getparam_behind_view_key, 
                               behind_view_cb );
    status |= add_keymap_entry( RACING, CONFIGURABLE_KEY, 
                               "x", getparam_follow_view_key, 
                               follow_view_cb );
    status |= add_keymap_entry( RACING, CONFIGURABLE_KEY, 
                               "c", getparam_above_view_key, above_view_cb );
    
    status |= add_keymap_entry( RACING, CONFIGURABLE_KEY, 
                               "=", getparam_screenshot_key, screenshot_cb );
    status |= add_keymap_entry( RACING, CONFIGURABLE_KEY, 
                               "p", getparam_pause_key, pause_cb );
    status |= add_keymap_entry( RACING, CONFIGURABLE_KEY, 
                               "r", getparam_reset_key, reset_cb );
    status |= add_keymap_entry( RACING, CONFIGURABLE_KEY, 
                               "space", getparam_jump_key, jump_cb );
    
    check_assertion( status == 0, "out of keymap entries" );
    
    register_loop_funcs( RACING, racing_init, racing_loop, racing_term );
}

void racing_mouse_func( int button, int state, int x, int y ) {
    static int noquit = 1;
    printf("racing_mouse_func %d %d %d\n", x, y, noquit);
    /* Check screen controls */
    paddling = False;
    braking = False;
    charging = False;
    if (state == SDL_PRESSED) {
        noquit = 1;
        if (x > 680 && y > 360)
            paddling = True;

        if (x < 120 && y >360)
            braking = True;

        if (x > 340 && x < 460 && y > 180 && y < 300) {
            noquit = 0;
            g_game.race_paused = True;
            set_game_mode( PAUSED );
        }

        if (x > 680 && y < 120)
            charging = True;

    }
    else if ( !noquit &&g_game.prev_mode == PAUSED && x < 120 && y > 360 && (state == SDL_RELEASED) ) {
        g_game.race_aborted = True;
        set_game_mode( GAME_OVER );
    }

    player_data_t *plyr = get_player_data( local_player() );
    if (paddling) plyr->control.is_accelerating = true; else  plyr->control.is_accelerating = false;
}

